/* tc-m68hc05.c -- Motorola 68HC05 assembler
   Copyright 2004 Free Software Foundation, Inc.
   Copyright 2004 Martin Kgler
   Contributed by Martin Kgler (mkoegler@auto.tuwien.ac.at)

   This file is part of GAS, the GNU Assembler.

   GAS is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GAS is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GAS; see the file COPYING.  If not, write to the Free
   Software Foundation, 51 Franklin Street - Fifth Floor, Boston, MA
   02110-1301, USA.  */

#include <stdio.h>
#include "dwarf2dbg.h"
#include "as.h"
#include "struc-symbol.h"

typedef enum
{
  M68HC05_NONE,
  M68HC05_IMM,
  M68HC05_OFFSET8,
  M68HC05_OFFSET16,
  M68HC05_INDEX0,
  M68HC05_INDEX1,
  M68HC05_INDEX2,
  M68HC05_PCREL,
  M68HC05_OFFSET8_PCREL,
} M68HC05_PARAM;

struct m68hc05_opcodes_s
{
  char *name;
  M68HC05_PARAM par;
  unsigned char opcode;
};

#define M68HC05_INS(NAME, PAR, OPCODE) \
{NAME, PAR, OPCODE},

static const struct m68hc05_opcodes_s m68hc05_opcodes[] = {
#include "opcode/m68hc05.h"
};

static const int m68hc05_opcnt =
  sizeof (m68hc05_opcodes) / sizeof (struct m68hc05_opcodes_s);

/*options*/
const char *md_shortopts = "";
struct option md_longopts[] = {
  {NULL, no_argument, NULL, 0}
};
size_t md_longopts_size = sizeof (md_longopts);

int
md_parse_option (int c ATTRIBUTE_UNUSED, char *arg ATTRIBUTE_UNUSED)
{
  return 0;
}

void
md_show_usage (FILE * stream ATTRIBUTE_UNUSED)
{
}


const pseudo_typeS md_pseudo_table[] = {
  {NULL, NULL, 0}
};

const char comment_chars[] = "#";
const char line_comment_chars[] = "#";
const char line_separator_chars[] = ";";

const char EXP_CHARS[] = "eE";
const char FLT_CHARS[] = "dD";


symbolS *
md_undefined_symbol (char ATTRIBUTE_UNUSED * name)
{
  return 0;
}


/* Turn a string in input_line_pointer into a floating point constant
   of type TYPE, and store the appropriate bytes in *LITP.  The number
   of LITTLENUMS emitted is stored in *SIZEP.  An error message is
   returned, or NULL on OK.  */

char *
md_atof (int type, char *litP, int *sizeP)
{
  return ieee_md_atof (type, litP, sizeP, TRUE);
}


static inline char *
skip_space (char *s)
{
  while (*s == ' ' || *s == '\t')
    ++s;
  return s;
}

/* Extract one word from FROM and copy it to TO.  */

static char *
extract_word (char *from, char *to, int limit)
{
  char *op_start;
  char *op_end;
  int size = 0;

  /* Drop leading whitespace.  */
  from = skip_space (from);
  *to = 0;

  /* Find the op code end.  */
  for (op_start = op_end = from; *op_end != 0 && is_part_of_name (*op_end);)
    {
      to[size++] = *op_end++;
      if (size + 1 >= limit)
	break;
    }

  to[size] = 0;
  return op_end;
}


static void m68hc05_set_relax_symbol (m68hc05_relax * rel, expressionS * exp,
				      m68hc05_op Op);


struct m68hc05_operation_s
{
  char *func;
  m68hc05_op operator;
};

struct m68hc05_operation_s m68hc05_operations[] = {
  {
   "lo8", O_LO8},
  {
   "hi8", O_HI8},
  {
   "hlo8", O_HLO8},
  {
   "hhi8", O_HHI8},
  {
   "hi16", O_HI16},
  {
   "lo16", O_LO16},
  {
   "offset8", O_OFFSET8},
  {
   "offset16", O_OFFSET16},
  {
   NULL, O_NONE},
};

static m68hc05_op
m68hc05_expression (expressionS * e)
{
  m68hc05_op o = O_NONE;
  struct m68hc05_operation_s *op = m68hc05_operations;
  while (op->func)
    {
      if (!strncmp (input_line_pointer, op->func, strlen (op->func)))
	{
	  char *str = input_line_pointer + strlen (op->func);
	  str = skip_space (str);
	  if (*str == '(')
	    {
	      o = op->operator;
	      input_line_pointer = skip_space (str + 1);
	      break;
	    }
	}
      op++;
    }
  expression (e);

  if (o != O_NONE)
    {
      input_line_pointer = skip_space (input_line_pointer);
      if (*input_line_pointer != ')')
	as_bad (_(") expected"));
      else
	input_line_pointer++;
    }
  return o;
}

static bfd_reloc_code_real_type
op2reloc (m68hc05_op o, int size)
{
  switch (size)
    {
    case 1:
      switch (o)
	{
	case O_LO8:
	  return BFD_RELOC_M68HC05_LO8;
	case O_HI8:
	  return BFD_RELOC_M68HC05_HI8;
	case O_HLO8:
	  return BFD_RELOC_M68HC05_HLO8;
	case O_HHI8:
	  return BFD_RELOC_M68HC05_HHI8;
	case O_OFFSET8:
	  return BFD_RELOC_M68HC05_SECTION_OFFSET8;
	case O_NONE:
	  return BFD_RELOC_M68HC05_RELAX_8;
	default:
	  as_bad (_("bad operator function"));
	  return BFD_RELOC_NONE;
	}
      break;
    case 2:
      switch (o)
	{
	case O_HI16:
	  return BFD_RELOC_HI16;
	case O_LO16:
	  return BFD_RELOC_LO16;
	case O_OFFSET16:
	  return BFD_RELOC_M68HC05_SECTION_OFFSET16;
	case O_NONE:
	  return BFD_RELOC_M68HC05_RELAX_16;
	default:
	  as_bad (_("bad operator function"));
	  return BFD_RELOC_NONE;
	}
      break;
    default:
      as_bad (_("bad operator function"));
      return BFD_RELOC_NONE;
    }
}
static bfd_reloc_code_real_type
op2overflow (m68hc05_op o)
{
  switch (o)
    {
    case O_LO8:
      return 1;
    case O_HI8:
      return 1;
    case O_HLO8:
      return 1;
    case O_HHI8:
      return 1;
    case O_HI16:
      return 1;
    case O_LO16:
      return 1;
    case O_NONE:
      return 0;
    case O_OFFSET8:
      return 0;
    case O_OFFSET16:
      return 0;
    default:
      as_bad (_("bad operator function"));
      return 0;
    }
}

static void
emit_none (unsigned char opcode)
{
  char *frag = frag_more (1);
  *frag = opcode;
}
static void
emit_X (unsigned char opcode)
{
  char *frag = frag_more (1);
  *frag = opcode;
}
static void
emit_imm (unsigned char opcode, m68hc05_op op, expressionS * ex)
{
  char *frag = frag_more (2);
  int where = frag - frag_now->fr_literal;
  fixS *fix;
  *frag++ = opcode;
  fix = fix_new_exp (frag_now, where + 1, 1, ex, 0, op2reloc (op, 1));
  fix->fx_no_overflow = op2overflow (op);
}
static void
emit_addr8 (unsigned char opcode, m68hc05_op op, expressionS * ex)
{
  char *frag = frag_more (2);
  int where = frag - frag_now->fr_literal;
  fixS *fix;
  *frag++ = opcode;
  fix = fix_new_exp (frag_now, where + 1, 1, ex, 0, op2reloc (op, 1));
  fix->fx_no_overflow = op2overflow (op);
}
static void
emit_index8 (unsigned char opcode, unsigned char opcode1 ATTRIBUTE_UNUSED,
	     m68hc05_op op, expressionS * ex)
{
  char *frag = frag_more (2);
  int where = frag - frag_now->fr_literal;
  fixS *fix;
  *frag++ = opcode;
  fix = fix_new_exp (frag_now, where + 1, 1, ex, 0, op2reloc (op, 1));
  fix->fx_no_overflow = op2overflow (op);
}
static void
emit_pcrel (unsigned char opcode, m68hc05_op op, expressionS * ex)
{
  if ((opcode & 0xf0) == 0x20)
    {
      m68hc05_relax *rel;
      frag_more (5);
      rel = &frag_now->tc_frag_data;
      frag_variant (rs_machine_dependent, 5, 2, 0, 0, 0, 0);
      m68hc05_set_relax_symbol (rel, ex, op);
      rel->opcode1 = opcode;
      rel->relax_type = REL_8_PCREL;
    }
  else
    {
      char *frag = frag_more (2);
      int where = frag - frag_now->fr_literal;
      fixS *fix;
      *frag++ = opcode;
      fix = fix_new_exp (frag_now, where + 1, 1, ex, 1, BFD_RELOC_8_PCREL);
    }
}
static void
emit_addr_pcrel (unsigned char opcode, m68hc05_op op1, expressionS * ex1,
		 m68hc05_op op2, expressionS * ex2)
{
  char *frag = frag_more (2);
  int where = frag - frag_now->fr_literal;
  fixS *fix;
  m68hc05_relax *rel;
  char *f = frag;
  *frag++ = opcode;
  fix = fix_new_exp (frag_now, where + 1, 1, ex1, 0, op2reloc (op1, 1));
  fix->fx_no_overflow = op2overflow (op1);
  frag_more (4);
  rel = &frag_now->tc_frag_data;
  frag_variant (rs_machine_dependent, 4, 1, 0, 0, 0, f);
  m68hc05_set_relax_symbol (rel, ex2, op2);
  rel->opcode1 = opcode;
  rel->relax_type = REL_ADDR_8_PCREL;

}

static void
m68hc05_set_relax_symbol (m68hc05_relax * rel, expressionS * exp,
			  m68hc05_op Op)
{
  symbolS *add = NULL;
  symbolS *sub = NULL;
  offsetT off = 0;

  switch (exp->X_op)
    {
    case O_absent:
      break;

    case O_register:
      as_bad (_("register value used as expression"));
      break;

    case O_add:
      /* This comes up when _GLOBAL_OFFSET_TABLE_+(.-L0) is read, if
         the difference expression cannot immediately be reduced.  */
      {
	symbolS *stmp = make_expr_symbol (exp);

	exp->X_op = O_symbol;
	exp->X_op_symbol = 0;
	exp->X_add_symbol = stmp;
	exp->X_add_number = 0;

	m68hc05_set_relax_symbol (rel, exp, Op);
	return;
      }

    case O_symbol_rva:
      add = exp->X_add_symbol;
      off = exp->X_add_number;

      break;

    case O_uminus:
      sub = exp->X_add_symbol;
      off = exp->X_add_number;
      break;

    case O_subtract:
      sub = exp->X_op_symbol;
      /* Fall through.  */
    case O_symbol:
      add = exp->X_add_symbol;
      /* Fall through.  */
    case O_constant:
      off = exp->X_add_number;
      break;

    default:
      add = make_expr_symbol (exp);
      break;
    }

  rel->reloc_info = Op;
  if (sub)
    as_bad (_("expression to complex"));
  rel->addsym = add;
  rel->off = off;

}


static void
emit_index16 (unsigned char opcode1, unsigned char opcode2,
	      unsigned char opcode3, m68hc05_op op, expressionS * ex)
{
  m68hc05_relax *rel;
  frag_more (3);
  rel = &frag_now->tc_frag_data;
  frag_variant (rs_machine_dependent, 3, 2, 0, 0, 0, 0);
  m68hc05_set_relax_symbol (rel, ex, op);
  rel->opcode1 = opcode1;
  rel->opcode2 = opcode2;
  rel->opcode3 = opcode3;
  rel->relax_type = REL_8_16_0;
}
static void
emit_addr16 (unsigned char opcode1, unsigned char opcode2, m68hc05_op op,
	     expressionS * ex)
{
  m68hc05_relax *rel;
  frag_more (3);
  rel = &frag_now->tc_frag_data;
  frag_variant (rs_machine_dependent, 3, 2, 0, 0, 0, 0);
  m68hc05_set_relax_symbol (rel, ex, op);
  rel->opcode1 = opcode1;
  rel->opcode2 = opcode2;
  rel->relax_type = REL_8_16;
}

static int
estimate_size (m68hc05_relax * rel)
{
  valueT val = 0;
  switch (rel->reloc_info)
    {
    case O_LO8:
      return 1;
    case O_HI8:
      return 1;
    case O_HLO8:
      return 1;
    case O_HHI8:
      return 1;
    case O_HI16:
      return 2;
    case O_LO16:
      return 2;
    case O_OFFSET8:
      return 1;
    case O_OFFSET16:
      return 2;
    case O_NONE:
      break;
    }
  if (rel->addsym)
    {
      val = resolve_symbol_value (rel->addsym);
      if (!S_IS_DEFINED (rel->addsym))
	return 2;
      return 2;			/*Sicherheitslsung */
    }
  val += rel->off;
  if (val == 0)
    return 0;
  if (val <= 0xff)
    return 1;
  return 2;
}

static int
estimate_size_8_PCREL (m68hc05_relax * rel, fragS * fragP)
{
  valueT val = 0;
  signed long relative;
  if (rel->addsym)
    {
      val = resolve_symbol_value (rel->addsym);
      if (!S_IS_DEFINED (rel->addsym))
	return 2;
    }
  val += rel->off;
  relative = val - (fragP->fr_address + fragP->fr_fix + 2);
  if (relative < 125 && relative > -125)
    return 1;
  return 2;
}

int
md_estimate_size_before_relax (fragS * fragp, asection * seg ATTRIBUTE_UNUSED)
{
  m68hc05_relax *rel = &fragp->tc_frag_data;
  int i;
  switch (rel->relax_type)
    {
    case REL_8_16:
      i = estimate_size (rel);
      if (i == 0)
	i = 1;
      return i + 1;
    case REL_8_16_0:
      return estimate_size (rel) + 1;
    case REL_8_PCREL:
      if (estimate_size_8_PCREL (rel, fragp) == 2)
	return 5;
      else
	return 2;
    case REL_ADDR_8_PCREL:
      if (estimate_size_8_PCREL (rel, fragp) == 2)
	return 4;
      else
	return 1;
    default:
      as_fatal (_("unknown relaxion type"));
    }
  return 0;
}

void
md_convert_frag (bfd * abfd ATTRIBUTE_UNUSED, asection * sec ATTRIBUTE_UNUSED,
		 fragS * fragP)
{
  char *pos = fragP->fr_fix + fragP->fr_literal;
  fixS *fix;
  m68hc05_relax *rel = &fragP->tc_frag_data;
  symbolS *sym;

  switch (rel->relax_type)
    {
    case REL_8_16:
      if (estimate_size (rel) <= 1)
	{
	  *pos++ = rel->opcode1;
	  fix =
	    fix_new (fragP, pos - fragP->fr_literal, 1, rel->addsym, rel->off,
		     0, op2reloc (rel->reloc_info, 1));
	  fix->fx_no_overflow = op2overflow (rel->reloc_info);
	  fragP->fr_fix += 2;
	}
      else
	{
	  *pos++ = rel->opcode2;
	  fix =
	    fix_new (fragP, pos - fragP->fr_literal, 2, rel->addsym, rel->off,
		     0, op2reloc (rel->reloc_info, 2));
	  fix->fx_no_overflow = op2overflow (rel->reloc_info);
	  fragP->fr_fix += 3;
	}
      break;
    case REL_8_16_0:
      if (estimate_size (rel) == 1)
	{
	  *pos++ = rel->opcode1;
	  fix =
	    fix_new (fragP, pos - fragP->fr_literal, 1, rel->addsym, rel->off,
		     0, op2reloc (rel->reloc_info, 1));
	  fix->fx_no_overflow = op2overflow (rel->reloc_info);
	  fragP->fr_fix += 2;
	}
      else if (estimate_size (rel) == 0)
	{
	  *pos++ = rel->opcode3;
	  fragP->fr_fix += 1;
	}
      else
	{
	  *pos++ = rel->opcode2;
	  fix =
	    fix_new (fragP, pos - fragP->fr_literal, 2, rel->addsym, rel->off,
		     0, op2reloc (rel->reloc_info, 2));
	  fix->fx_no_overflow = op2overflow (rel->reloc_info);
	  fragP->fr_fix += 3;
	}
      break;
    case REL_ADDR_8_PCREL:
      if (estimate_size_8_PCREL (rel, fragP) == 2)
	{
	  sym = symbol_temp_new (sec, fragP->fr_fix + 4, fragP);
	  fragP->fr_opcode[0] = rel->opcode1 ^ 0x1;
	  fix_new (fragP, pos - fragP->fr_literal - 2, 4, rel->addsym,
		   rel->off, 0, BFD_RELOC_M68HC05_RELAX_GROUP);
	  fix =
	    fix_new (fragP, pos - fragP->fr_literal, 1, sym, 0,
		     1, BFD_RELOC_8_PCREL);
	  fix->fx_no_overflow = 0;
	  *pos++ = 0;
	  *pos++ = 0xCC;
	   /*JMP*/
	    fix =
	    fix_new (fragP, pos - fragP->fr_literal, 2, rel->addsym, rel->off,
		     0, BFD_RELOC_M68HC05_RELAX_16);
	  fix->fx_no_overflow = 0;
	  fragP->fr_fix += 4;
	}
      else
	{
	  fragP->fr_opcode[0] = rel->opcode1;
	  fix =
	    fix_new (fragP, pos - fragP->fr_literal, 1, rel->addsym, rel->off,
		     1, BFD_RELOC_8_PCREL);
	  fix->fx_no_overflow = 0;
	  fragP->fr_fix += 1;
	}
      break;
    case REL_8_PCREL:
      if (estimate_size_8_PCREL (rel, fragP) == 2)
	{
	  sym = symbol_temp_new (sec, fragP->fr_fix + 5, fragP);
	  fix_new (fragP, pos - fragP->fr_literal, 4, rel->addsym, rel->off,
		   0, BFD_RELOC_M68HC05_RELAX_GROUP);
	  *pos++ = rel->opcode1 ^ 1;
	  fix =
	    fix_new (fragP, pos - fragP->fr_literal, 1, sym, 0,
		     1, BFD_RELOC_8_PCREL);
	  fix->fx_no_overflow = 0;
	  *pos++ = 0;
	  *pos++ = 0xCC;
	   /*JMP*/
	    fix =
	    fix_new (fragP, pos - fragP->fr_literal, 2, rel->addsym, rel->off,
		     0, BFD_RELOC_M68HC05_RELAX_16);
	  fix->fx_no_overflow = 0;
	  fragP->fr_fix += 5;
	}
      else
	{
	  *pos++ = rel->opcode1;
	  fix =
	    fix_new (fragP, pos - fragP->fr_literal, 1, rel->addsym, rel->off,
		     1, BFD_RELOC_M68HC05_RELAX_PCREL_8);
	  fix->fx_no_overflow = 0;
	  fragP->fr_fix += 2;
	}
      break;
    default:
      as_fatal (_("unknown relaxion type"));
    }
}

int next_rewriteid = 0;

/* change names ending in !!!!! in an uniq name for the file */
char *
m68hc05_rewrite_name (char *name)
{
  int len = strlen (name);
  if (len >= 5 && !strcmp (name + len - 5, "!!!!!"))
    {
      char buf[6];
      snprintf (buf, 6, "%05X", next_rewriteid++);
      memcpy (name + len - 5, buf, 5);
      return name;
    }
  return name;
}

void
md_begin ()
{
}

void
md_assemble (char *str)
{
  const struct m68hc05_opcodes_s *ops = m68hc05_opcodes;
  char op[11];
  char *savep = input_line_pointer;
  int i, i1, i2;
  expressionS e1, e2;
  m68hc05_op o1 = 0, o2 = 0;
  enum
  { PARNONE, PARIMM, PARINDEX0, PARINDEX1, PARADDR, PAR2ADDR }
  parmode;
  int done = 0;

  str = skip_space (extract_word (str, op, sizeof (op)));

  if (!op[0])
    {
      as_bad (_("can't find opcode "));
      return;
    }

  i1 = i2 = -1;
  for (i = 0; i < m68hc05_opcnt; i++)
    if (!strcmp (ops[i].name, op))
      {
	i2 = i;
	if (i1 == -1)
	  i1 = i;
      }
  if (i1 == -1)
    {
      as_bad (_("unknown opcode `%s'"), op);
      return;
    }


  str = skip_space (str);
  parmode = PARNONE;
  if (*str == '%' && *(str + 1) == 'X'
      && (*(str + 2) == ' ' || *(str + 2) == '\t' || *(str + 2) == ','
	  || !*(str + 2)))
    {
      str += 2;
      parmode = PARINDEX0;
      str = skip_space (str);
      if (*str == ',')
	{
	  parmode = PARINDEX1;
	  str++;
	  input_line_pointer = str;
	  o1 = m68hc05_expression (&e1);
	  str = input_line_pointer;
	}
    }
  else if (*str == '$')
    {
      parmode = PARIMM;
      str++;
      str = skip_space (str);
      input_line_pointer = str;
      o1 = m68hc05_expression (&e1);
      str = input_line_pointer;
    }
  else if (*str)
    {
      parmode = PARADDR;
      input_line_pointer = str;
      o1 = m68hc05_expression (&e1);
      str = input_line_pointer;
      str = skip_space (str);
      if (*str == ',')
	{
	  str++;
	  parmode = PAR2ADDR;
	  str = skip_space (str);
	  input_line_pointer = str;
	  o2 = m68hc05_expression (&e2);
	  str = input_line_pointer;
	}
    }

  if (*skip_space (str))
    as_bad (_("garbage at end of line"));
  input_line_pointer = savep;

  dwarf2_emit_insn (0);

  switch (parmode)
    {
    case PARNONE:
      if (i1 != i2)
	as_bad (_("internal error"));
      if (ops[i1].par != M68HC05_NONE)
	as_bad (_("no parameter expected"));
      emit_none (ops[i1].opcode);
      done = 1;
      break;
    case PARIMM:
      for (i = i1; i <= i2; i++)
	if (ops[i].par == M68HC05_IMM)
	  {
	    emit_imm (ops[i].opcode, o1, &e1);
	    done = 1;
	    break;
	  }
      break;
    case PARINDEX0:
      for (i = i1; i <= i2; i++)
	if (ops[i].par == M68HC05_INDEX0)
	  {
	    emit_X (ops[i].opcode);
	    done = 1;
	    break;
	  }
      break;
    case PARINDEX1:
      for (i = i1; i <= i2; i++)
	{
	  if (ops[i].par == M68HC05_INDEX1 && ops[i + 1].par == M68HC05_INDEX2
	      && ops[i + 2].par == M68HC05_INDEX0 && i + 1 < i2)
	    {
	      emit_index16 (ops[i].opcode, ops[i + 1].opcode,
			    ops[i + 2].opcode, o1, &e1);
	      done = 1;
	      break;
	    }
	  else
	    if (ops[i].par == M68HC05_INDEX1
		&& ops[i + 1].par == M68HC05_INDEX0 && i < i2)
	    {
	      emit_index8 (ops[i].opcode, ops[i + 1].opcode, o1, &e1);
	      done = 1;
	      break;
	    }
	}
      break;
    case PARADDR:
      for (i = i1; i <= i2; i++)
	{
	  if (ops[i].par == M68HC05_OFFSET8
	      && ops[i + 1].par == M68HC05_OFFSET16 && i < i2)
	    {
	      emit_addr16 (ops[i].opcode, ops[i + 1].opcode, o1, &e1);
	      done = 1;
	      break;
	    }
	  else if (ops[i].par == M68HC05_OFFSET8)
	    {
	      emit_addr8 (ops[i].opcode, o1, &e1);
	      done = 1;
	      break;
	    }
	  else if (ops[i].par == M68HC05_PCREL)
	    {
	      emit_pcrel (ops[i].opcode, o1, &e1);
	      done = 1;
	      break;
	    }
	}
      break;
    case PAR2ADDR:
      for (i = i1; i <= i2; i++)
	if (ops[i].par == M68HC05_OFFSET8_PCREL)
	  {
	    emit_addr_pcrel (ops[i].opcode, o1, &e1, o2, &e2);
	    done = 1;
	    break;
	  }
      break;
    default:
      as_bad (_("internal error"));
    }
  if (!done)
    as_bad (_("unexpected parameter type"));
}



/* GAS will call this function for each section at the end of the assembly,
   to permit the CPU backend to adjust the alignment of a section.  */

valueT
md_section_align (asection * seg, valueT addr)
{
  int align = bfd_get_section_alignment (stdoutput, seg);
  return ((addr + (1 << align) - 1) & (-1 << align));
}

/* A `BFD_ASSEMBLER' GAS will call this to generate a reloc.  GAS
   will pass the resulting reloc to `bfd_install_relocation'.  This
   currently works poorly, as `bfd_install_relocation' often does the
   wrong thing, and instances of `tc_gen_reloc' have been written to
   work around the problems, which in turns makes it difficult to fix
   `bfd_install_relocation'.  */

/* If while processing a fixup, a reloc really needs to be created
   then it is done here.  */

arelent *
tc_gen_reloc (asection * seg ATTRIBUTE_UNUSED, fixS * fixp)
{
  arelent *reloc;

  reloc = (arelent *) xmalloc (sizeof (arelent));

  reloc->sym_ptr_ptr = (asymbol **) xmalloc (sizeof (asymbol *));
  *reloc->sym_ptr_ptr = symbol_get_bfdsym (fixp->fx_addsy);

  reloc->address = fixp->fx_frag->fr_address + fixp->fx_where;
  reloc->howto = bfd_reloc_type_lookup (stdoutput, fixp->fx_r_type);
  if (reloc->howto == (reloc_howto_type *) NULL)
    {
      as_bad_where (fixp->fx_file, fixp->fx_line,
		    _("reloc %d not supported by object file format"),
		    (int) fixp->fx_r_type);
      return NULL;
    }

  if (fixp->fx_r_type == BFD_RELOC_VTABLE_INHERIT
      || fixp->fx_r_type == BFD_RELOC_VTABLE_ENTRY)
    reloc->address = fixp->fx_offset;

  reloc->addend = fixp->fx_offset;

  return reloc;
}


/* If you define this macro, it should return the offset between the
   address of a PC relative fixup and the position from which the PC
   relative adjustment should be made.  On many processors, the base
   of a PC relative instruction is the next instruction, so this
   macro would return the length of an instruction.  */

long
md_pcrel_from_section (fixS * fixp, segT sec)
{
  if (fixp->fx_addsy != (symbolS *) NULL
      && (!S_IS_DEFINED (fixp->fx_addsy)
	  || (S_GET_SEGMENT (fixp->fx_addsy) != sec)))
    return 0;

  return fixp->fx_frag->fr_address + fixp->fx_where + fixp->fx_size;
}

/* GAS will call this for each fixup.  It should store the correct
   value in the object file.  */

void
md_apply_fix (fixS * fixP, valueT * valP, segT seg)
{
  unsigned char *where;
  long value = *valP;

  if (fixP->fx_addsy == (symbolS *) NULL)
    fixP->fx_done = 1;
  else if (fixP->fx_addsy && fixP->fx_addsy->sy_resolved &&
	   fixP->fx_addsy->bsym && 
	   fixP->fx_addsy->bsym->section == absolute_section)
    fixP->fx_done = 1;
  else if (fixP->fx_pcrel)
    {
      segT s = S_GET_SEGMENT (fixP->fx_addsy);

      if (s == seg || s == absolute_section)
	{
	  value += S_GET_VALUE (fixP->fx_addsy);
	  fixP->fx_done = 0;	/* 1; */
	}
    }

  /* can only be resolved at link time */
  if (fixP->fx_r_type == BFD_RELOC_M68HC05_SECTION_OFFSET8)
    fixP->fx_done = 0;
  if (fixP->fx_r_type == BFD_RELOC_M68HC05_SECTION_OFFSET16)
    fixP->fx_done = 0;

  /* We don't actually support subtracting a symbol.  */
  if (fixP->fx_subsy != (symbolS *) NULL)
    as_bad_where (fixP->fx_file, fixP->fx_line, _("expression too complex"));

  if (fixP->fx_done)
    {
      where = (unsigned char *) fixP->fx_frag->fr_literal + fixP->fx_where;
      switch (fixP->fx_r_type)
	{
	case BFD_RELOC_M68HC05_LO8:
	  *where = (value >> 0) & 0xff;
	  break;
	case BFD_RELOC_M68HC05_HI8:
	  *where = (value >> 8) & 0xff;
	  break;
	case BFD_RELOC_M68HC05_HLO8:
	  *where = (value >> 16) & 0xff;
	  break;
	case BFD_RELOC_M68HC05_HHI8:
	  *where = (value >> 24) & 0xff;
	  break;
	case BFD_RELOC_8:
	  *where = (value) & 0xff;
	  break;
	case BFD_RELOC_16:
	  *where = (value >> 8) & 0xff;
	  where++;
	  *where = (value >> 0) & 0xff;
	  break;
	case BFD_RELOC_M68HC05_RELAX_8:
	  *where = (value) & 0xff;
	  break;
	case BFD_RELOC_M68HC05_RELAX_16:
	  *where = (value >> 8) & 0xff;
	  where++;
	  *where = (value >> 0) & 0xff;
	  break;
	case BFD_RELOC_32:
	  *where = (value >> 24) & 0xff;
	  where++;
	  *where = (value >> 16) & 0xff;
	  where++;
	  *where = (value >> 8) & 0xff;
	  where++;
	  *where = (value >> 0) & 0xff;
	  break;
	case BFD_RELOC_LO16:
	  *where = (value >> 8) & 0xff;
	  where++;
	  *where = (value >> 0) & 0xff;
	  break;
	case BFD_RELOC_HI16:
	  *where = (value >> 24) & 0xff;
	  where++;
	  *where = (value >> 16) & 0xff;
	  break;
	case BFD_RELOC_M68HC05_RELAX_PCREL_8:
	case BFD_RELOC_8_PCREL:
	  *where = (value) & 0xff;
	  break;
	case BFD_RELOC_M68HC05_RELAX_GROUP:
	  break;
	default:
	  as_bad_where (fixP->fx_file, fixP->fx_line,
			_("unknown relocation"));
	}

    }
}
int
tc_m68hc05_force_relocation (struct fix *fixP)
{
  return 1 || generic_force_reloc (fixP);
}

int
tc_m68hc05_fix_adjustable (struct fix *fixP ATTRIBUTE_UNUSED)
{
  return 0;
}

/* variable to pass lo8/hi8/.. for m68hc05_parse_cons_expression */
static m68hc05_op cons_mod = 0;

void
m68hc05_parse_cons_expression (expressionS * exp, int nbytes ATTRIBUTE_UNUSED)
{

  cons_mod = m68hc05_expression (exp);
}

void
m68hc05_cons_fix_new (fragS * frag, int where, int nbytes, expressionS * exp)
{
  switch (nbytes)
    {
    case 1:
      switch (cons_mod)
	{
	case O_NONE:
	  fix_new_exp (frag, where, nbytes, exp, FALSE, BFD_RELOC_8);
	  break;

	case O_LO8:
	  fix_new_exp (frag, where, nbytes, exp, FALSE,
		       BFD_RELOC_M68HC05_LO8);
	  break;

	case O_HI8:
	  fix_new_exp (frag, where, nbytes, exp, FALSE,
		       BFD_RELOC_M68HC05_HI8);
	  break;

	case O_HLO8:
	  fix_new_exp (frag, where, nbytes, exp, FALSE,
		       BFD_RELOC_M68HC05_HLO8);
	  break;

	case O_HHI8:
	  fix_new_exp (frag, where, nbytes, exp, FALSE,
		       BFD_RELOC_M68HC05_HHI8);
	  break;

	case O_OFFSET8:
	  fix_new_exp (frag, where, nbytes, exp, FALSE,
		       BFD_RELOC_M68HC05_SECTION_OFFSET8);
	  break;

	default:
	  as_bad (_("unsupported relocation for size: %d"), nbytes);
	}
      return;

    case 2:
      switch (cons_mod)
	{
	case O_NONE:
	  fix_new_exp (frag, where, nbytes, exp, FALSE, BFD_RELOC_16);
	  break;

	case O_LO16:
	  fix_new_exp (frag, where, nbytes, exp, FALSE, BFD_RELOC_LO16);
	  break;

	case O_HI16:
	  fix_new_exp (frag, where, nbytes, exp, FALSE, BFD_RELOC_HI16);
	  break;

	case O_OFFSET16:
	  fix_new_exp (frag, where, nbytes, exp, FALSE,
		       BFD_RELOC_M68HC05_SECTION_OFFSET16);
	  break;

	default:
	  as_bad (_("unsupported relocation for size: %d"), nbytes);
	}
      return;

    case 4:
      switch (cons_mod)
	{
	case O_NONE:
	  fix_new_exp (frag, where, nbytes, exp, FALSE, BFD_RELOC_32);
	  break;
	default:
	  as_bad (_("unsupported relocation for size: %d"), nbytes);
	}
      return;

    default:
      as_bad (_("illegal relocation size: %d"), nbytes);
    }
}
